/*
ID: espr1t
TASK: parash
KEYWORDS: NP, Hill-Climbing, Greedy, Brute-Force
CONTEST: CodeIT 2012, Round 1
*/

#include <cstdio>
#include <cstring>
#include <ctime>
#include <algorithm>
#include <vector>

#define DEBUG_INFO 0
#define TIME_CHECKS 10
#define MAX_EXEC_TIME 1.5
#define MAX_CRIT_TIME 1.7
#define MAX_BACKTRACK_SIZE 100
#define MAX_BACKTRACK_ANIMALS 20
#define MAX_RANDOM_ITERATIONS 100

using namespace std;

const int INF = 1000000001;
const int MAX_ANIMALS = 104;
const int MAX_BOARD_SIZE = 54;
const int MAX_ANIMAL_SIZE = 16;
const int MAX_ANIMAL_LEGS = 104;

struct Answer {
    int index;
    int atRow, atCol;
    Answer(int index_ = 0, int atRow_ = 0, int atCol_ = 0) :
        index(index_), atRow(atRow_), atCol(atCol_) {}
    bool operator < (const Answer& r) const {
        if (index != r.index)
            return index > r.index;
        return atRow != r.atRow ? atRow < r.atRow : atCol < r.atCol;
    }
};

struct Animal {
    int index;
    int reduceFactor, safetyLimit;
    int numRows, numCols;
    int numLegs;
    int legs[MAX_ANIMAL_LEGS][2];
    vector <Answer> positions;
    
    // All comparators return "true" if they have to be swapped
    static bool compareByArea(const Animal& a1, const Animal& a2) {
        return a1.numRows * a1.numCols < a2.numRows * a2.numCols;
    }
    static bool compareByLegs(const Animal& a1, const Animal& a2) {
        return a1.numLegs < a2.numLegs;
    }
    static bool compareByReduceFactor(const Animal& a1, const Animal& a2) {
        return a1.reduceFactor > a2.reduceFactor;
    }
    static bool compareBySafetyLimit(const Animal& a1, const Animal& a2) {
        return a1.safetyLimit < a2.safetyLimit;
    }
    static bool compareBySLtoRF(const Animal& a1, const Animal& a2) {
        return (double)a1.safetyLimit / a1.reduceFactor <
               (double)a2.safetyLimit / a2.reduceFactor;
    }
    static bool compareByRFtoSL(const Animal& a1, const Animal& a2) {
        return (double)a1.reduceFactor / a1.safetyLimit <
               (double)a2.reduceFactor / a2.safetyLimit;
    }
};

double lastTime;
unsigned sTime, timeCalls;

int numRows, numCols;
int board[MAX_BOARD_SIZE][MAX_BOARD_SIZE];
int initialBoard[MAX_BOARD_SIZE][MAX_BOARD_SIZE];

int numAnimals;
Animal animals[MAX_ANIMALS];

int bestCount, bestAnswer;
Answer answer[MAX_ANIMALS];

int curCount, curAnswer;
Answer current[MAX_ANIMALS];

inline double getTime() {
    if (timeCalls++ % TIME_CHECKS == 0) {
        lastTime = (double)(clock() - sTime) / CLOCKS_PER_SEC;
    }
    return lastTime;
}

void bubbleSort(bool (*comparator)(const Animal&, const Animal&)) {    
    // We'll implement a sophisticated sorting algorithm called "Bubble Sort"
    for (int i = 0; i < numAnimals; i++)
        for (int c = 0; c + 1 < numAnimals; c++)
            if (comparator(animals[c], animals[c + 1]))
                swap(animals[c], animals[c + 1]);
}

inline int putAtPos(Animal& animal, int atRow, int atCol, bool updateBoard) {
    if (atRow < 0 || atRow + animal.numRows > numRows) return -INF;
    if (atCol < 0 || atCol + animal.numCols > numCols) return -INF;

    int ret = 0;
    for (int i = 0; i < animal.numLegs; i++) {
        int row = atRow + animal.legs[i][0];
        int col = atCol + animal.legs[i][1];
        if (board[row][col] < animal.safetyLimit)
            return -INF;
        ret += board[row][col];
        if (updateBoard)
            board[row][col] /= animal.reduceFactor;
    }
    return ret;
}

void improveMore() {
    int dirs[8][2] = { {-1, -1}, {-1, 0}, {-1, +1}, {0, +1},
                       {+1, +1}, {+1, 0}, {+1, -1}, {0, -1} };
    
    curAnswer = bestAnswer; curCount = bestCount;
    for (int i = 0; i < curCount; i++)
        current[i] = answer[i];
    
    int rev[MAX_ANIMALS];
    for (int i = 0; i < numAnimals; i++)
        rev[animals[i].index] = i;

    // This is somewhat odd. Trying to do an "earthquake" - moving pairs of
    // landing locations a bit to the sides and see if that leads to improvement
    while (true) {
        bool flag = false;
        for (int idx1 = 0; idx1 < curCount; idx1++) {
            for (int idx2 = idx1 + 1; idx2 < curCount; idx2++) {
                for (int i = 0; i < 8; i++) for (int m1 = 1; m1 < 3; m1++) {
                    for (int c = 0; c < 8; c++) for (int m2 = 1; m2 < 3; m2++) {
                        if (getTime() > MAX_CRIT_TIME)
                            break;
                        current[idx1].atRow += dirs[i][0] * m1;
                        current[idx1].atCol += dirs[i][1] * m1;
                        current[idx2].atRow += dirs[c][0] * m2;
                        current[idx2].atCol += dirs[c][1] * m2;

                        int currentShuffle = 0;
                        memcpy(board, initialBoard, sizeof(board));
                        for (int who = 0; who < curCount; who++) {
                            currentShuffle += putAtPos(animals[rev[current[who].index]],
                                    current[who].atRow, current[who].atCol, true);
                            if (currentShuffle < 0)
                                break;
                        }

                        if (curAnswer < currentShuffle) {
                            curAnswer = currentShuffle;
                            flag = true;
                        }
                        else {
                            current[idx1].atRow -= dirs[i][0] * m1;
                            current[idx1].atCol -= dirs[i][1] * m1;
                            current[idx2].atRow -= dirs[c][0] * m2;
                            current[idx2].atCol -= dirs[c][1] * m2;
                        }
                    }
                    if (getTime() > MAX_CRIT_TIME)
                        break;
                }
                if (getTime() > MAX_CRIT_TIME)
                    break;
            }
            if (getTime() > MAX_CRIT_TIME)
                break;
        }
        if (getTime() > MAX_CRIT_TIME)
            break;
        if (!flag) break;
    }
    if (bestAnswer < curAnswer) {
        bestCount = curCount;
        bestAnswer = curAnswer;
        for (int i = 0; i < curCount; i++)
            answer[i] = current[i];
    }
}

// Greedy positioning of animals
// Once the order in which they jump is set, just find the best place for
// each of them and put it there (potentially screwing some after that)
int greedy() {
    curCount = 0; curAnswer = 0;
    memcpy(board, initialBoard, sizeof(board));
    
    for (int i = 0; i < numAnimals; i++) {
        // Find best position
        Animal& animal = animals[i];
        int atRow = -1, atCol = -1, win = 0;
        for (int c = 0; c < (int)animal.positions.size(); c++) {
            if (animal.positions[c].index <= win)
                break;
            int curWin = putAtPos(animal,
                animal.positions[c].atRow, animal.positions[c].atCol, false);
            if (win < curWin) {
                win = curWin;
                atRow = animal.positions[c].atRow;
                atCol = animal.positions[c].atCol;
            }
        }
        if (atRow != -1 && atCol != -1) {
            current[curCount].index = animals[i].index;
            current[curCount].atRow = atRow;
            current[curCount].atCol = atCol;
            curCount++;
            curAnswer += putAtPos(animal, atRow, atCol, true);
        }
    }
    
    // Heeeeey, sexy ladyyyy
    if (bestAnswer < curAnswer) {
        bestCount = curCount;
        bestAnswer = curAnswer;
        for (int i = 0; i < curCount; i++)
            answer[i] = current[i];
    }
    return curAnswer;
}

// Evaluate function
// Uses hill-climbing to set the best order of animals
// Swaps the order of pairs of animals, going in the direction of the first
// successful swap. Note that this is faster than the usual hill-climbing,
// however also results in (usually) finding not as good local optimum.
void eval() {
    int currentBest = greedy();
    if (getTime() > MAX_EXEC_TIME)
        return;

    while (true) {
        bool flag = false;
        for (int i = 0; i < numAnimals - 1; i++) {
            for (int c = i + 1; c < numAnimals; c++) {
                swap(animals[i], animals[c]);
                int swapResult = greedy();
                if (currentBest < swapResult)
                    currentBest = swapResult, flag = true;
                else
                    swap(animals[i], animals[c]);
                if (getTime() > MAX_EXEC_TIME)
                    break;
            }
            if (getTime() > MAX_EXEC_TIME)
                {flag = false; break;}
        }
        if (!flag) break;
    }
}

// Brute force for small inputs
// Really bad implementation -- it is incredibly slow.
// Right now I'm sick and decided just to have *some* backtrack,
// because Vladi was winning on the small tests.
int recCnt;
int save[MAX_ANIMALS][MAX_BOARD_SIZE][MAX_BOARD_SIZE];

bool recurse(int idx) {
    if (idx >= numAnimals) {
        if (bestAnswer < curAnswer) {
            bestAnswer = curAnswer, bestCount = curCount;
            for (int i = 0; i < curCount; i++)
                answer[i] = current[i];
            if (DEBUG_INFO)
                fprintf(stderr, "  >> found new best with backtrack!\n");
        }
        return true;
    }
    
    // Resistance is futile! (Hitchhiker's Guide to the Galaxy)
    // Check if it is pointless to continue
    // This, although quite slow, actually prunes the search tree a lot
    int maxAns = curAnswer;
    for (int i = idx; i < numAnimals; i++) {
        Animal& animal = animals[i];
        int best = 0;
        for (int c = 0; c < (int)animal.positions.size(); c++) {
            int atRow = animal.positions[c].atRow;
            int atCol = animal.positions[c].atCol;
            best = max(best, putAtPos(animal, atRow, atCol, false));
        }
        maxAns += best;
    }
    if (maxAns <= bestAnswer)
        return true;
    
    // Try every single position
    Animal& animal = animals[idx];
    for (int row = 0; row < numRows; row++)
        for (int col = 0; col < numCols; col++)
            save[idx][row][col] = board[row][col];
    
    for (int i = 0; i < (int)animal.positions.size(); i++) {
        int atRow = animal.positions[i].atRow;
        int atCol = animal.positions[i].atCol;
        int add = putAtPos(animal, atRow, atCol, true);

        bool timeOkay = true;
        if (add >= 0) {
            current[curCount] = Answer(animal.index, atRow, atCol);
            curAnswer += add; curCount++;
            timeOkay = recurse(idx + 1);
            curAnswer -= add; curCount--;
        }
        for (int c = 0; c < animal.numLegs; c++) {
            int row = atRow + animal.legs[c][0];
            int col = atCol + animal.legs[c][1];
            board[row][col] = save[idx][row][col];
        }
        if (!timeOkay)
            return false;
        recCnt++;
        if (recCnt % 10 == 0 && getTime() > MAX_EXEC_TIME)
            return false;
    }
    return true;
}

// Op... op, op, oppa Gangnam style!
// I couldn't think of any appropriate function name.
void gangnamStyle() {
    // Since probably the most important thing for this task is the order
    // of animals to jump, we'll try several possible arrangements.

    // By Area
    if (getTime() < MAX_EXEC_TIME) {
        bubbleSort(&Animal::compareByArea);
        eval(); reverse(animals, animals + numAnimals); eval();
    }
    
    // By Number of Legs
    if (getTime() < MAX_EXEC_TIME) {
        bubbleSort(&Animal::compareByLegs);
        eval(); reverse(animals, animals + numAnimals); eval();
    }

    // By Safety Limit
    if (getTime() < MAX_EXEC_TIME) {
        bubbleSort(&Animal::compareBySafetyLimit);
        eval(); reverse(animals, animals + numAnimals); eval();
    }
    
    // By Reduce Factor
    if (getTime() < MAX_EXEC_TIME) {
        bubbleSort(&Animal::compareByReduceFactor);
        eval(); reverse(animals, animals + numAnimals); eval();
    }
    
    // By Safety Limit / Reduce Factor
    if (getTime() < MAX_EXEC_TIME) {
        bubbleSort(&Animal::compareBySLtoRF);
        eval(); reverse(animals, animals + numAnimals); eval();
    }
    
    // By Reduce Factor / Safety Limit
    if (getTime() < MAX_EXEC_TIME) {
        bubbleSort(&Animal::compareByRFtoSL);
        eval(); reverse(animals, animals + numAnimals); eval();
    }
    
    // And then we'll try several more...
    int numRandoms = 0;
    // Random Shuffles Style
    while (getTime() < MAX_EXEC_TIME) {
        if (++numRandoms >= MAX_RANDOM_ITERATIONS)
            break;
        random_shuffle(animals, animals + numAnimals);
        eval();
    }
    
    // Still have time and board is small?
    if (numRows * numCols <= MAX_BACKTRACK_SIZE && numAnimals <= MAX_BACKTRACK_ANIMALS) {
        srand(42);
        int numBacktracks = 0;
        while (getTime() < MAX_EXEC_TIME) {
            numBacktracks++;
            // First backtrack is special!
            if (numBacktracks == 1)
                bubbleSort(&Animal::compareByArea);
            else
                random_shuffle(animals, animals + numAnimals);
            memcpy(board, initialBoard, sizeof(board));
            curAnswer = 0; curCount = 0; recCnt = 0;
            recurse(0);
        }
        if (DEBUG_INFO)
            fprintf(stderr, "  -- ran %d backtracks.\n", numBacktracks);
    }
    
    // Some more randoms, because we can still have time
    // (test is small, but not *that* small to run backtrack)
    while (getTime() < MAX_EXEC_TIME) {
        numRandoms++;
        random_shuffle(animals, animals + numAnimals);
        eval();
    }
    if (DEBUG_INFO)
        fprintf(stderr, "  -- number of random iterations: %d\n", numRandoms);

    // Relying on this function to improve the result is almost as
    // optimistic as calling Jessica Alba and asking her out "just in case".
    improveMore();
}

// This is quite nice optimization,
// it improves the runtime of the greedy by several times
// Basically just iterate the placing positions of an animal and cut
// the search once we're sure we cannot find a better place
void precalcPositions() {
    memcpy(board, initialBoard, sizeof(board));
    for (int idx = 0; idx < numAnimals; idx++) {
        Animal& animal = animals[idx];
        Answer add;
        for (add.atRow = 0; add.atRow + animal.numRows <= numRows; add.atRow++) {
            for (add.atCol = 0; add.atCol + animal.numCols <= numCols; add.atCol++) {
                add.index = putAtPos(animals[idx], add.atRow, add.atCol, false);
                if (add.index > 0)
                    animal.positions.push_back(add);
            }
        }
        sort(animal.positions.begin(), animal.positions.end());
    }
}

void readInput(FILE* in) {
    fscanf(in, "%d %d %d", &numRows, &numCols, &numAnimals);
    for (int row = 0; row < numRows; row++)
        for (int col = 0; col < numCols; col++)
            fscanf(in, "%d", &initialBoard[row][col]);
    
    for (int i = 0; i < numAnimals; i++) {
        Animal& animal = animals[i];
        animal.index = i;
        fscanf(in, "%d %d", &animal.numRows, &animal.numCols);
        fscanf(in, "%d %d", &animal.reduceFactor, &animal.safetyLimit);
        // The random generator creates animals with 30-40% less "legs" than
        // actual area. Use this as a speed improvement.
        animal.numLegs = 0;
        for (int row = 0; row < animal.numRows; row++) {
            char buffer[MAX_ANIMAL_SIZE];
            fscanf(in, "%s", buffer);
            for (int col = 0; col < animal.numCols; col++) {
                if (buffer[col] == '1') {
                    animal.legs[animal.numLegs][0] = row;
                    animal.legs[animal.numLegs][1] = col;
                    animal.numLegs++;
                }
            }
        }
    }
}

void printOutput(FILE* out) {
    if (DEBUG_INFO) {
        fprintf(stderr, "  -- best answer: %d\n", bestAnswer);
        fprintf(stderr, "  -- num clock() calls: %u\n", timeCalls);
    }
//    fprintf(out, "%d\n", bestAnswer);
    fprintf(out, "%d\n", bestCount);
    for (int i = 0; i < bestCount; i++) {
        fprintf(out, "%d %d %d\n",
                answer[i].index + 1, answer[i].atRow + 1, answer[i].atCol + 1);
    }
}

void init() {
    // Random Generator
    srand(42);
    
    // Time handling
    sTime = clock();
    timeCalls = 0; lastTime = 0.0;
    
    // Answer
    curCount = curAnswer = 0;
    bestCount = bestAnswer = 0;

    // Variables
    for (int i = 0; i < MAX_ANIMALS; i++)
        animals[i].positions.clear();
}

int main(void) {
    // This can be made more elegant, but was easy to test.
    int numTests = 1;
    FILE* in = fopen("parash.in", "rt");
    FILE* out = fopen("parash.out", "wt");

    double maxExecTime = 0.0;
    for (int test = 0; test < numTests; test++) {
        if (DEBUG_INFO)
            fprintf(stderr, "Testcase %d out of %d...\n", test + 1, numTests);

        init();
        readInput(in);
        precalcPositions();
        gangnamStyle();
        printOutput(out);
        maxExecTime = max(maxExecTime, getTime());
        
        if (DEBUG_INFO)
            fprintf(stderr, "  -- execution time: %.3lf\n", getTime());
    }
    if (DEBUG_INFO)
        fprintf(stderr, "Maximal execution time was: %.3lf\n", maxExecTime);
	return 0;
}
